# -*- coding: utf-8 -*-
# Copyright 2018, Red Hat, Inc.
# License: GPL-2.0+ <http://spdx.org/licenses/GPL-2.0+>
# See the LICENSE file for more details on Licensing

'''Functional tests for libtaskotron/executor.py'''

import pytest
import mock
import signal

from libtaskotron import executor
import libtaskotron.exceptions as exc
from libtaskotron import config

PLAYBOOK='''
- hosts: localhost
  # this saves a lot of time when running in mock (without network)
  gather_facts: no
  vars:
    taskotron_generic_task: true
  tasks:
    - debug:
        msg: This is a sample debug printout
'''


@pytest.mark.usefixtures('setup')
class TestExecutor():

    @pytest.fixture
    def setup(self, tmpdir, monkeypatch):
        '''Run this before every test invocation'''
        self.artifactsdir = tmpdir.mkdir('artifacts')
        self.taskdir = tmpdir.mkdir('taskdir')
        self.client_taskdir = tmpdir.mkdir('client_taskdir')
        self.arg_data = {
            'artifactsdir': self.artifactsdir.strpath,
            'taskdir': self.taskdir.strpath,
            'item': 'htop-2.0.2-4.fc27',
            'type': 'koji_build',
            'arch': 'noarch',
            'debug': False,
            'ssh_privkey': None,
        }
        self.playbook_name = 'tests.yml'
        self.playbook = self.taskdir.join(self.playbook_name)
        self.playbook.write(PLAYBOOK)
        self.ipaddr = '127.0.0.1'
        self.executor = executor.Executor(self.arg_data)

        monkeypatch.setattr(config, '_config', None)
        self.conf = config.get_config()
        self.conf.client_taskdir = self.client_taskdir.strpath

    def test_syntax_check_valid(self):
        '''Should not raise if playbook has valid syntax'''
        self.executor._check_playbook_syntax(self.playbook.strpath)

    def test_syntax_check_invalid(self):
        '''Should raise if playbook has invalid syntax'''
        self.playbook.write("This surely isn't a valid Ansible playbook")
        with pytest.raises(exc.TaskotronPlaybookError):
            self.executor._check_playbook_syntax(self.playbook.strpath)

    def test_run_playbook_simple(self, monkeypatch):
        '''Execute a very simple playbook whether everything works'''
        mock_signal = mock.Mock()
        monkeypatch.setattr(signal, 'signal', mock_signal)

        # the tradeoff here is that we need to run with root=False
        # (non-default)
        output = self.executor._run_playbook(self.playbook_name, self.ipaddr,
            root=False)
        # FIXME: We currently have no idea whether the test playbook passed
        # or failed, because we ignore the inner playbook's exit status. So
        # we can't really verify here whether everything worked ok.
        # To work around this, we grep for known lines in stdout. The warning
        # printout about failed task must be skipped.
        lines = output.splitlines()
        matched = [line for line in lines
            if line.startswith('TASK [Warn about failed task]')]
        assert len(matched) == 1
        warn_task_line = lines.index(matched[0])
        assert 'skipping' in lines[warn_task_line + 1]
