# -*- coding: utf-8 -*-
# Copyright 2018, Red Hat, Inc.
# License: GPL-2.0+ <http://spdx.org/licenses/GPL-2.0+>
# See the LICENSE file for more details on Licensing

'''Unit tests for libtaskotron/executor.py'''

import pytest
import mock
import signal
import os
import json
import subprocess

from libtaskotron import executor
import libtaskotron.exceptions as exc
from libtaskotron.directives import resultsdb_directive
from libtaskotron import os_utils

PLAYBOOK='''
- hosts: localhost
  # this saves a lot of time when running in mock (without network)
  gather_facts: no
  vars:
    taskotron_generic_task: true
  tasks:
    - debug:
        msg: This is a sample debug printout
'''


@pytest.mark.usefixtures('setup')
class TestExecutor():

    @pytest.fixture
    def setup(self, tmpdir):
        '''Run this before every test invocation'''
        self.artifactsdir = tmpdir.mkdir('artifacts')
        self.taskdir = tmpdir.mkdir('taskdir')
        self.arg_data = {
            'artifactsdir': self.artifactsdir.strpath,
            'taskdir': self.taskdir.strpath,
            'item': 'htop-2.0.2-4.fc27',
            'type': 'koji_build',
            'arch': 'noarch',
            'debug': False,
            'local': True,
            'libvirt': False,
            'ssh': False,
            'ssh_privkey': None,
            'uuid': 'random-uuid',
        }
        self.playbook_name = 'tests.yml'
        self.playbook = self.taskdir.join(self.playbook_name)
        self.playbook.write(PLAYBOOK)
        self.ipaddr = '127.0.0.1'
        self.executor = executor.Executor(self.arg_data)

    def test_interrupt_handler(self):
        '''Raise correct exception for an interrupt'''

        signals = {
            1: 'SIGHUP',
            2: 'SIGINT',
            3: 'SIGQUIT',
            15: 'SIGTERM',
            99: 'UNKNOWN',
        }

        for (signum, signame) in signals.items():
            try:
                self.executor._interrupt_handler(signum, None)
            except exc.TaskotronInterruptError as e:
                assert e.signum == signum
                assert e.signame == signame
                assert str(e.signum) in str(e)
                assert e.signame in str(e)

    def test_report_results(self, monkeypatch):
        '''Report results using resultsdb if results file exists'''
        playbook = 'playbook.yml'
        resultsfile = self.artifactsdir.mkdir(playbook).mkdir(
            'taskotron').join('results.yml')
        resultsfile.ensure()

        mock_rdb = mock.Mock()
        monkeypatch.setattr(resultsdb_directive, 'ResultsdbDirective',
            mock_rdb)
        rdb_instance = mock_rdb.return_value

        self.executor._report_results(playbook)
        mock_rdb.assert_called_once()  # create instance
        rdb_instance.process.assert_called_once()
        assert rdb_instance.process.call_args[1]['params']['file'] == \
            resultsfile.strpath
        assert rdb_instance.process.call_args[1]['arg_data'] == \
            self.arg_data

    def test_report_results_missing_file(self, monkeypatch):
        '''Don't try to report results when no results file exists'''
        mock_rdb = mock.Mock()
        monkeypatch.setattr(resultsdb_directive, 'ResultsdbDirective',
            mock_rdb)

        self.executor._report_results('missing-playbook-dir')
        mock_rdb.assert_not_called()

    def test_run_playbook(self, monkeypatch):
        '''A standard invocation of ansible-playbook'''
        mock_check_syntax = mock.Mock()
        monkeypatch.setattr(executor.Executor, '_check_playbook_syntax',
            mock_check_syntax)
        mock_signal = mock.Mock()
        monkeypatch.setattr(signal, 'signal', mock_signal)
        mock_popen = mock.Mock(return_value=('fake output', None))
        monkeypatch.setattr(os_utils, 'popen_rt', mock_popen)

        output = self.executor._run_playbook(self.playbook_name, self.ipaddr)

        # must return playbook output
        assert output == 'fake output'

        # must check syntax
        mock_check_syntax.assert_called_once()

        # must mask signals
        assert mock_signal.call_count == 4  # 2 signals masked, then reset

        # must export ansible vars
        varsfile = os.path.join(self.artifactsdir.strpath, self.playbook_name,
            'taskotron', 'ansible_vars.json')
        varsfile_internal = os.path.join(self.artifactsdir.strpath,
            self.playbook_name, 'taskotron', 'ansible_vars_internal.json')
        for vf in [varsfile, varsfile_internal]:
            assert os.path.isfile(vf)
            with open(vf, 'r') as f:
                json.load(f)

        # must run ansible-playbook
        mock_popen.assert_called_once()
        cmd_args = mock_popen.call_args[0][0]
        assert cmd_args[0] == 'ansible-playbook'
        assert '--inventory={},'.format(self.ipaddr) in cmd_args
        assert '--extra-vars=@{}'.format(varsfile) in cmd_args
        assert '--extra-vars=@{}'.format(varsfile_internal) in cmd_args
        assert '--become' in cmd_args
        assert '--connection=local' in cmd_args

    def test_run_playbook_failed(self, monkeypatch):
        '''Should raise when ansible-playbook fails'''
        mock_check_syntax = mock.Mock()
        monkeypatch.setattr(executor.Executor, '_check_playbook_syntax',
            mock_check_syntax)
        mock_signal = mock.Mock()
        monkeypatch.setattr(signal, 'signal', mock_signal)
        cpe = subprocess.CalledProcessError(returncode=99, cmd='fake')
        mock_popen = mock.Mock(side_effect=cpe)
        monkeypatch.setattr(os_utils, 'popen_rt', mock_popen)

        with pytest.raises(exc.TaskotronError):
            self.executor._run_playbook(self.playbook_name, self.ipaddr)

        # must unmask signals even when playbook failed
        assert mock_signal.call_count == 4  # 2 signals masked, then reset

        # must run ansible-playbook
        mock_popen.assert_called_once()

    def test_run_playbook_interrupted(self, monkeypatch):
        '''Should try failsafe stop when interrupted'''
        mock_check_syntax = mock.Mock()
        monkeypatch.setattr(executor.Executor, '_check_playbook_syntax',
            mock_check_syntax)
        mock_signal = mock.Mock()
        monkeypatch.setattr(signal, 'signal', mock_signal)
        error = exc.TaskotronInterruptError(15, 'SIGTERM')
        mock_popen = mock.Mock(side_effect=[error, mock.DEFAULT])
        monkeypatch.setattr(os_utils, 'popen_rt', mock_popen)

        with pytest.raises(exc.TaskotronInterruptError):
            self.executor._run_playbook(self.playbook_name, self.ipaddr)

        # must unmask signals even when playbook failed
        assert mock_signal.call_count == 4  # 2 signals masked, then reset

        # must run ansible-playbook twice, fist normal, second failsafe
        assert mock_popen.call_count == 2
        assert '--tags' in mock_popen.call_args_list[1][0][0]
        assert 'failsafe' in mock_popen.call_args_list[1][0][0]

    def test_run_playbook_failsafe_error(self, monkeypatch):
        '''When failsafe stop gives an error, it should be ignored'''
        mock_check_syntax = mock.Mock()
        monkeypatch.setattr(executor.Executor, '_check_playbook_syntax',
            mock_check_syntax)
        mock_signal = mock.Mock()
        monkeypatch.setattr(signal, 'signal', mock_signal)
        error = exc.TaskotronInterruptError(15, 'SIGTERM')
        cpe = subprocess.CalledProcessError(returncode=99, cmd='fake')
        mock_popen = mock.Mock(side_effect=[error, cpe])
        monkeypatch.setattr(os_utils, 'popen_rt', mock_popen)

        with pytest.raises(exc.TaskotronInterruptError):
            self.executor._run_playbook(self.playbook_name, self.ipaddr)

    def test_execute_local(self, monkeypatch):
        '''Execution using local mode'''
        mock_spawn_vm = mock.Mock()
        monkeypatch.setattr(executor.Executor, '_spawn_vm', mock_spawn_vm)
        mock_run_playbook = mock.Mock()
        monkeypatch.setattr(executor.Executor, '_run_playbook',
            mock_run_playbook)
        mock_report_results = mock.Mock()
        monkeypatch.setattr(executor.Executor, '_report_results',
            mock_report_results)
        assert self.executor.arg_data['local'] == True

        finished = self.executor.execute()

        assert finished == True
        mock_spawn_vm.assert_not_called()
        mock_run_playbook.assert_called_once()
        assert mock_run_playbook.call_args[0][1] == '127.0.0.1'
        mock_report_results.assert_called_once()

    def test_execute_ssh(self, monkeypatch):
        '''Execution using ssh mode'''
        mock_spawn_vm = mock.Mock()
        monkeypatch.setattr(executor.Executor, '_spawn_vm', mock_spawn_vm)
        mock_run_playbook = mock.Mock()
        monkeypatch.setattr(executor.Executor, '_run_playbook',
            mock_run_playbook)
        mock_report_results = mock.Mock()
        monkeypatch.setattr(executor.Executor, '_report_results',
            mock_report_results)
        self.executor.arg_data['local'] = False
        self.executor.arg_data['ssh'] = True
        self.executor.arg_data['machine'] = '127.0.0.2'

        finished = self.executor.execute()

        assert finished == True
        mock_spawn_vm.assert_not_called()
        mock_run_playbook.assert_called_once()
        assert mock_run_playbook.call_args[0][1] == '127.0.0.2'
        mock_report_results.assert_called_once()

    def test_execute_libvirt(self, monkeypatch):
        '''Execution using libvirt mode'''
        mock_spawn_vm = mock.Mock(return_value='127.0.0.3')
        monkeypatch.setattr(executor.Executor, '_spawn_vm', mock_spawn_vm)
        mock_run_playbook = mock.Mock()
        monkeypatch.setattr(executor.Executor, '_run_playbook',
            mock_run_playbook)
        mock_report_results = mock.Mock()
        monkeypatch.setattr(executor.Executor, '_report_results',
            mock_report_results)
        mock_task_vm = mock.Mock()
        self.executor.task_vm = mock_task_vm
        self.executor.arg_data['local'] = False
        self.executor.arg_data['libvirt'] = True

        finished = self.executor.execute()

        assert finished == True
        mock_spawn_vm.assert_called_once()
        mock_run_playbook.assert_called_once()
        assert mock_run_playbook.call_args[0][1] == '127.0.0.3'
        mock_report_results.assert_called_once()
        mock_task_vm.teardown.assert_called_once()

    def test_execute_no_playbooks(self, monkeypatch):
        '''Should raise when there are no playbooks'''
        mock_run_playbook = mock.Mock()
        monkeypatch.setattr(executor.Executor, '_run_playbook',
            mock_run_playbook)
        mock_report_results = mock.Mock()
        monkeypatch.setattr(executor.Executor, '_report_results',
            mock_report_results)
        self.playbook.remove()

        with pytest.raises(exc.TaskotronError):
            self.executor.execute()

        mock_run_playbook.assert_not_called()
        mock_report_results.assert_not_called()

    def test_execute_more_playbooks(self, monkeypatch):
        '''Should execute all found playbooks'''
        mock_run_playbook = mock.Mock()
        monkeypatch.setattr(executor.Executor, '_run_playbook',
            mock_run_playbook)
        mock_report_results = mock.Mock()
        monkeypatch.setattr(executor.Executor, '_report_results',
            mock_report_results)
        self.playbook.copy(self.taskdir.join('tests_copy.yml'))

        finished = self.executor.execute()

        assert finished == True
        assert mock_run_playbook.call_count == 2
        playbooks = [
            mock_run_playbook.call_args_list[0][0][0],
            mock_run_playbook.call_args_list[1][0][0]
        ]
        assert 'tests.yml' in playbooks
        assert 'tests_copy.yml' in playbooks
        assert mock_report_results.call_count == 2
        playbooks = [
            mock_report_results.call_args_list[0][0][0],
            mock_report_results.call_args_list[1][0][0]
        ]
        assert 'tests.yml' in playbooks
        assert 'tests_copy.yml' in playbooks

    def test_execute_error(self, monkeypatch):
        '''Should ignore playbook errors'''
        mock_run_playbook = mock.Mock(side_effect=exc.TaskotronError)
        monkeypatch.setattr(executor.Executor, '_run_playbook',
            mock_run_playbook)
        mock_report_results = mock.Mock()
        monkeypatch.setattr(executor.Executor, '_report_results',
            mock_report_results)

        finished = self.executor.execute()

        assert finished == True
        mock_run_playbook.assert_called_once()
        mock_report_results.assert_not_called()

    def test_execute_interrupted(self, monkeypatch):
        '''Should halt execution and return when interrupted'''
        error = exc.TaskotronInterruptError(15, 'SIGTERM')
        mock_run_playbook = mock.Mock(side_effect=error)
        monkeypatch.setattr(executor.Executor, '_run_playbook',
            mock_run_playbook)
        mock_report_results = mock.Mock()
        monkeypatch.setattr(executor.Executor, '_report_results',
            mock_report_results)
        self.playbook.copy(self.taskdir.join('tests_copy.yml'))

        finished = self.executor.execute()

        assert finished == False
        mock_run_playbook.assert_called_once()
        mock_report_results.assert_not_called()
